'''
Copyright (c) 2019-2021 TIV Consultancy. All right reserved

No redistribution in any form allowed.
'''

from mxnet.gluon import nn
from mxnet import nd

class ResNetBlock(nn.HybridBlock):
    def __init__(self, chn, **kwargs):
        super(ResNetBlock, self).__init__(**kwargs)
        self.chn = chn
        with self.name_scope():
            self.bn1 = nn.BatchNorm()
            self.conv1 = nn.Conv2D(in_channels=self.chn, channels=self.chn, kernel_size=3, strides=1, padding=1, dilation=1)
            self.bn2 = nn.BatchNorm()
            self.conv2 = nn.Conv2D(in_channels=self.chn, channels=self.chn, kernel_size=3, strides=1, padding=1, dilation=1)

    def hybrid_forward(self, F, x):
        res = x
        x = self.bn1(x)
        x = F.Activation(x, act_type='relu')
        x = self.conv1(x)
        x = self.bn2(x)
        x = F.Activation(x, act_type='relu')
        x = self.conv2(x)
        return F.concat(res,x, dim=1)

class StartBlockDownSample(nn.HybridBlock):
    def __init__(self, **kwargs):
        super(StartBlockDownSample, self).__init__(**kwargs)
        with self.name_scope():
            self.conv1 = nn.HybridSequential(prefix='start_conv1')
            self.conv2 = nn.HybridSequential(prefix='start_conv2')

            self.conv1.add(nn.Conv2D(in_channels=1, channels=64, kernel_size=3, strides=1, padding=1, dilation=1))
            self.conv1.add(nn.BatchNorm())
            self.conv2.add(nn.Conv2D(in_channels=64, channels=64, kernel_size=3, strides=1, padding=1, dilation=1))
            self.conv2.add(nn.BatchNorm())

    def hybrid_forward(self, F, x):
        x = self.conv1(x)
        x = F.Activation(x, act_type='relu')
        x = self.conv2(x)
        x = F.Activation(x, act_type='relu')
        return x


class StartBlockDownSampleDilation(nn.HybridBlock):
    def __init__(self, dilatation, **kwargs):
        super(StartBlockDownSampleDilation, self).__init__(**kwargs)
        with self.name_scope():
            self.conv1 = nn.HybridSequential(prefix='start_conv1')
            self.conv2 = nn.HybridSequential(prefix='start_conv2')

            self.conv1.add(nn.Conv2D(in_channels=1, channels=64, kernel_size=3, strides=1, padding=dilatation,
                                     dilation=dilatation))
            self.conv1.add(nn.BatchNorm())
            self.conv2.add(nn.Conv2D(in_channels=64, channels=64, kernel_size=3, strides=1, padding=dilatation,
                                     dilation=dilatation))
            self.conv2.add(nn.BatchNorm())

    def hybrid_forward(self, F, x):
        x = self.conv1(x)
        x = F.Activation(x, act_type='relu')
        x = self.conv2(x)
        x = F.Activation(x, act_type='relu')
        return x


class BlockDownSample(nn.HybridBlock):
    def __init__(self, in_channels, **kwargs):
        super(BlockDownSample, self).__init__(**kwargs)
        with self.name_scope():
            self.conv1 = nn.HybridSequential(prefix='down_conv1' + str(in_channels))
            self.conv2 = nn.HybridSequential(prefix='down_conv2' + str(in_channels))
            self.max_pool = nn.MaxPool2D(pool_size=2, strides=2, padding=0)

            self.conv1.add(nn.Conv2D(in_channels=in_channels, channels=in_channels * 2, kernel_size=3, strides=1,
                                     padding=1, dilation=1))
            self.conv1.add(nn.BatchNorm())
            self.conv2.add(nn.Conv2D(in_channels=in_channels * 2, channels=in_channels * 2, kernel_size=3, strides=1,
                                     padding=1, dilation=1))
            self.conv2.add(nn.BatchNorm())

    def hybrid_forward(self, F, x):
        x = self.max_pool(x)
        x = self.conv1(x)
        x = F.Activation(x, act_type='relu')
        x = self.conv2(x)
        x = F.Activation(x, act_type='relu')
        return x

class BlockDownSampleRes(nn.HybridBlock):
    def __init__(self, in_channels, **kwargs):
        super(BlockDownSampleRes, self).__init__(**kwargs)
        with self.name_scope():
            self.res = ResNetBlock(chn=in_channels)
            self.max_pool = nn.MaxPool2D(pool_size=2, strides=2, padding=0)

    def hybrid_forward(self, F, x):
        x = self.max_pool(x)
        x = self.res(x)
        return x


class BlockDownSampleDilatation(nn.HybridBlock):
    def __init__(self, dilatation, in_channels, **kwargs):
        super(BlockDownSampleDilatation, self).__init__(**kwargs)
        with self.name_scope():
            self.conv1 = nn.HybridSequential(prefix='down_conv1' + str(in_channels))
            self.conv2 = nn.HybridSequential(prefix='down_conv2' + str(in_channels))

            self.max_pool = nn.MaxPool2D(pool_size=2, strides=2, padding=0)
            self.conv1.add(nn.Conv2D(in_channels=in_channels, channels=in_channels * 2, kernel_size=3, strides=1,
                                     padding=dilatation, dilation=dilatation))
            self.conv1.add(nn.BatchNorm())
            self.conv2.add(nn.Conv2D(in_channels=in_channels * 2, channels=in_channels * 2, kernel_size=3, strides=1,
                                     padding=dilatation, dilation=dilatation))
            self.conv2.add(nn.BatchNorm())

    def hybrid_forward(self, F, x):
        x = self.max_pool(x)
        x = self.conv1(x)
        x = F.Activation(x, act_type='relu')
        x = self.conv2(x)
        x = F.Activation(x, act_type='relu')
        return x


class BaseBlock(nn.HybridBlock):
    def __init__(self, in_channels, **kwargs):
        super(BaseBlock, self).__init__(**kwargs)
        with self.name_scope():
            self.conv1 = nn.HybridSequential(prefix='base_conv1' + str(in_channels))
            self.conv2 = nn.HybridSequential(prefix='base_conv2' + str(in_channels))
            self.up_sample = nn.HybridSequential(prefix='base_up' + str(in_channels))

            self.max_pool = nn.MaxPool2D(pool_size=2, strides=2, padding=0)
            self.conv1.add(nn.Conv2D(in_channels=in_channels, channels=in_channels * 2, kernel_size=3, strides=1,
                                     padding=1, dilation=1))
            self.conv1.add(nn.BatchNorm())
            self.conv2.add(nn.Conv2D(in_channels=in_channels * 2, channels=in_channels * 2, kernel_size=3, strides=1,
                                     padding=1, dilation=1))
            self.conv2.add(nn.BatchNorm())
            self.up_sample.add(nn.Conv2DTranspose(in_channels=in_channels * 2, channels=in_channels, kernel_size=2,
                                                  padding=0, strides=2))

    def hybrid_forward(self, F, x):
        x = self.max_pool(x)
        x = self.conv1(x)
        x = F.Activation(x, act_type='relu')
        x = self.conv2(x)
        x = F.Activation(x, act_type='relu')
        x = self.up_sample(x)
        return x


class BaseBlockDilation(nn.HybridBlock):
    def __init__(self, in_channels, dilation, **kwargs):
        super(BaseBlockDilation, self).__init__(**kwargs)
        with self.name_scope():
            self.conv1 = nn.HybridSequential(prefix='base_conv1' + str(in_channels))
            self.conv2 = nn.HybridSequential(prefix='base_conv2' + str(in_channels))
            self.up_sample = nn.HybridSequential(prefix='base_up' + str(in_channels))

            self.max_pool = nn.MaxPool2D(pool_size=2, strides=2, padding=0)
            self.conv1.add(nn.Conv2D(in_channels=in_channels, channels=in_channels * 2, kernel_size=3, strides=1,
                                     padding=dilation, dilation=dilation))
            self.conv1.add(nn.BatchNorm())
            self.conv2.add(nn.Conv2D(in_channels=in_channels * 2, channels=in_channels * 2, kernel_size=3, strides=1,
                                     padding=dilation, dilation=dilation))
            self.conv2.add(nn.BatchNorm())
            self.up_sample.add(nn.Conv2DTranspose(in_channels=in_channels * 2, channels=in_channels, kernel_size=2,
                                                  padding=0, strides=2))

    def hybrid_forward(self, F, x):
        x = self.max_pool(x)
        x = self.conv1(x)
        x = F.Activation(x, act_type='relu')
        x = self.conv2(x)
        x = F.Activation(x, act_type='relu')
        x = self.up_sample(x)
        return x

class BlockUpSample(nn.HybridBlock):
    def __init__(self, in_channels, **kwargs):
        super(BlockUpSample, self).__init__(**kwargs)
        with self.name_scope():
            self.conv1 = nn.HybridSequential(prefix='up_conv1' + str(in_channels))
            self.conv2 = nn.HybridSequential(prefix='up_conv2' + str(in_channels))
            self.up_sample = nn.HybridSequential(prefix='up_sample' + str(in_channels))

            self.conv1.add(nn.Conv2D(in_channels=in_channels*2, channels=in_channels, kernel_size=3, strides=1,
                                     padding=1, dilation=1))
            self.conv1.add(nn.BatchNorm())
            self.conv2.add(nn.Conv2D(in_channels=in_channels, channels=in_channels, kernel_size=3, strides=1,
                                     padding=1, dilation=1))
            self.conv2.add(nn.BatchNorm())
            self.up_sample.add(nn.Conv2DTranspose(in_channels=in_channels, channels=in_channels//2, kernel_size=2,
                                                  padding=0, strides=2))

    def hybrid_forward(self, F, x, cross):
        x = F.concat(x, cross, dim=1)
        x = self.conv1(x)
        x = F.Activation(x, act_type='relu')
        x = self.conv2(x)
        x = F.Activation(x, act_type='relu')
        x = self.up_sample(x)
        return x


class BlockUpSampleDilatation(nn.HybridBlock):
    def __init__(self, in_channels, dilation, **kwargs):
        super(BlockUpSampleDilatation, self).__init__(**kwargs)
        with self.name_scope():
            self.conv1 = nn.HybridSequential(prefix='up_conv1' + str(in_channels))
            self.conv2 = nn.HybridSequential(prefix='up_conv2' + str(in_channels))
            self.up_sample = nn.HybridSequential(prefix='up_sample' + str(in_channels))

            self.conv1.add(nn.Conv2D(in_channels=in_channels*2, channels=in_channels, kernel_size=3, strides=1,
                                     padding=dilation, dilation=dilation))
            self.conv1.add(nn.BatchNorm())
            self.conv2.add(nn.Conv2D(in_channels=in_channels, channels=in_channels, kernel_size=3, strides=1,
                                     padding=dilation, dilation=dilation))
            self.conv2.add(nn.BatchNorm())
            self.up_sample.add(nn.Conv2DTranspose(in_channels=in_channels, channels=in_channels//2, kernel_size=2,
                                                  padding=0, strides=2))

    def hybrid_forward(self, F, x, cross):
        x = F.concat(x, cross, dim=1)
        x = self.conv1(x)
        x = F.Activation(x, act_type='relu')
        x = self.conv2(x)
        x = F.Activation(x, act_type='relu')
        x = self.up_sample(x)
        return x


class BlockUpDilatationSample(nn.HybridBlock):
    def __init__(self, in_channels, **kwargs):
        super(BlockUpDilatationSample, self).__init__(**kwargs)
        with self.name_scope():
            self.conv1 = nn.HybridSequential(prefix='up_conv1' + str(in_channels))
            self.conv2 = nn.HybridSequential(prefix='up_conv2' + str(in_channels))
            self.up_sample = nn.HybridSequential(prefix='up_sample' + str(in_channels))

            self.conv1.add(nn.Conv2D(in_channels=in_channels*2, channels=in_channels, kernel_size=3, strides=1,
                                     padding=1, dilation=1))
            self.conv1.add(nn.BatchNorm())
            self.conv2.add(nn.Conv2D(in_channels=in_channels, channels=in_channels, kernel_size=3, strides=1,
                                     padding=1, dilation=1))
            self.conv2.add(nn.BatchNorm())
            self.up_sample.add(nn.Conv2DTranspose(in_channels=in_channels, channels=in_channels//2, kernel_size=2,
                                                  padding=1, strides=2, dilation=3))

    def hybrid_forward(self, F, x, cross):
        x = nd.concat(x, cross, dim=1)
        x = self.conv1(x)
        x = F.Activation(x, act_type='relu')
        x = self.conv2(x)
        x = F.Activation(x, act_type='relu')
        x = self.up_sample(x)
        return x


class EndBlock(nn.HybridBlock):
    def __init__(self, in_channels, **kwargs):
        super(EndBlock, self).__init__(**kwargs)
        with self.name_scope():
            self.conv1 = nn.HybridSequential(prefix='up_conv1' + str(in_channels))
            self.conv2 = nn.HybridSequential(prefix='up_conv2' + str(in_channels))

            self.conv1.add(nn.Conv2D(in_channels=in_channels*2, channels=in_channels, kernel_size=3, strides=1,
                                     padding=1, dilation=1))
            self.conv1.add(nn.BatchNorm())
            self.conv2.add(nn.Conv2D(in_channels=in_channels, channels=in_channels, kernel_size=3, strides=1,
                                     padding=1, dilation=1))
            self.conv2.add(nn.BatchNorm())

    def hybrid_forward(self, F, x, cross):
        x = F.concat(x, cross, dim=1)
        x = self.conv1(x)
        x = F.Activation(x, act_type='relu')
        x = self.conv2(x)
        x = F.Activation(x, act_type='relu')
        return x

class EndBlockDilatation(nn.HybridBlock):
    def __init__(self, in_channels, dilatation, **kwargs):
        super(EndBlockDilatation, self).__init__(**kwargs)
        with self.name_scope():
            self.conv1 = nn.HybridSequential(prefix='up_conv1' + str(in_channels))
            self.conv2 = nn.HybridSequential(prefix='up_conv2' + str(in_channels))

            self.conv1.add(nn.Conv2D(in_channels=in_channels*2, channels=in_channels, kernel_size=3, strides=1,
                                     padding=dilatation, dilation=dilatation))
            self.conv1.add(nn.BatchNorm())
            self.conv2.add(nn.Conv2D(in_channels=in_channels, channels=in_channels, kernel_size=3, strides=1,
                                     padding=dilatation, dilation=dilatation))
            self.conv2.add(nn.BatchNorm())

    def hybrid_forward(self, F, x, cross):
        x = F.concat(x, cross, dim=1)
        x = self.conv1(x)
        x = F.Activation(x, act_type='relu')
        x = self.conv2(x)
        x = F.Activation(x, act_type='relu')
        return x


class BlockClassifier(nn.HybridBlock):
    def __init__(self, in_channels, classes, **kwargs):
        super(BlockClassifier, self).__init__(**kwargs)
        with self.name_scope():
            self.conv1 = nn.HybridSequential(prefix='class_conv1' + str(in_channels))

            self.conv1.add(nn.Conv2D(in_channels=in_channels, channels=classes, kernel_size=1, strides=1,
                                     padding=0, dilation=1))

    def hybrid_forward(self, F, x):
        x = self.conv1(x)
        return x


class UNetL1(nn.HybridBlock):
    r"""UNetL1
    Parameters
    ----------
    classes : int, default 2
        Number of classification classes.
    """

    def __init__(self, classes=2, **kwargs):
        super(UNetL1, self).__init__(**kwargs)
        with self.name_scope():
            self.start = StartBlockDownSample()
            self.basement = BaseBlock(64)
            self.end = EndBlock(64)
            self.classifier = BlockClassifier(64, classes)

    def hybrid_forward(self, F, x):
        x = self.start(x)
        cross1 = x
        x = self.basement(x)
        x = self.end(x, cross1)
        x = self.classifier(x)
        return x

class UNetL3(nn.HybridBlock):
    r"""UNetL3
    Parameters
    ----------
    classes : int, default 2
        Number of classification classes.
    """

    def __init__(self, classes=2, **kwargs):
        super(UNetL3, self).__init__(**kwargs)
        with self.name_scope():
            self.start = StartBlockDownSample()
            self.down1 = BlockDownSample(64)
            self.down2 = BlockDownSample(128)
            self.basement = BaseBlock(256)
            self.up1 = BlockUpSample(256)
            self.up2 = BlockUpSample(128)
            self.end = EndBlock(64)
            self.classifier = BlockClassifier(64, classes)

    def hybrid_forward(self, F, x):
        x = self.start(x)
        cross1 = x
        x = self.down1(x)
        cross2 = x
        x = self.down2(x)
        cross3 = x
        x = self.basement(x)
        x = self.up1(x, cross3)
        x = self.up2(x, cross2)
        x = self.end(x, cross1)
        x = self.classifier(x)
        return x


class UNetL5(nn.HybridBlock):
    r"""UNetL5
    Parameters
    ----------
    classes : int, default 2
        Number of classification classes.
    """

    def __init__(self, classes=2, **kwargs):
        super(UNetL5, self).__init__(**kwargs)
        with self.name_scope():
            self.start = StartBlockDownSample()
            self.down1 = BlockDownSample(64)
            self.down2 = BlockDownSample(128)
            self.down3 = BlockDownSample(256)
            self.down4 = BlockDownSample(512)
            self.basement = BaseBlock(1024)
            self.up1 = BlockUpSample(1024)
            self.up2 = BlockUpSample(512)
            self.up3 = BlockUpSample(256)
            self.up4 = BlockUpSample(128)
            self.end = EndBlock(64)
            self.classifier = BlockClassifier(64, classes)

    def hybrid_forward(self, F, x):
        x = self.start(x)
        cross1 = x
        x = self.down1(x)
        cross2 = x
        x = self.down2(x)
        cross3 = x
        x = self.down3(x)
        cross4 = x
        x = self.down4(x)
        cross5 = x
        x = self.basement(x)
        x = self.up1(x, cross5)
        x = self.up2(x, cross4)
        x = self.up3(x, cross3)
        x = self.up4(x, cross2)
        x = self.end(x, cross1)
        x = self.classifier(x)
        return x
    
class UNetL5Dilation(nn.HybridBlock):
    r"""UNetL5
    Parameters
    ----------
    classes : int, default 2
        Number of classification classes.
    """

    def __init__(self, classes=2, **kwargs):
        super(UNetL5Dilation, self).__init__(**kwargs)
        with self.name_scope():
            self.start = StartBlockDownSampleDilation(3)
            self.down1 = BlockDownSampleDilatation(3,64)
            self.down2 = BlockDownSampleDilatation(3,128)
            self.down3 = BlockDownSampleDilatation(3,256)
            self.down4 = BlockDownSampleDilatation(3,512)
            self.basement = BaseBlockDilation(1024,3)
            self.up1 = BlockUpSampleDilatation(1024,3)
            self.up2 = BlockUpSampleDilatation(512,3)
            self.up3 = BlockUpSampleDilatation(256,3)
            self.up4 = BlockUpSampleDilatation(128,3)
            self.end = EndBlockDilatation(64,3)
            self.classifier = BlockClassifier(64, classes)

    def hybrid_forward(self, F, x):
        x = self.start(x)
        cross1 = x
        x = self.down1(x)
        cross2 = x
        x = self.down2(x)
        cross3 = x
        x = self.down3(x)
        cross4 = x
        x = self.down4(x)
        cross5 = x
        x = self.basement(x)
        x = self.up1(x, cross5)
        x = self.up2(x, cross4)
        x = self.up3(x, cross3)
        x = self.up4(x, cross2)
        x = self.end(x, cross1)
        x = self.classifier(x)
        return x