'''
Copyright (c) 2019-2021 TIV Consultancy. All right reserved

No redistribution in any form allowed.
'''

import os
import random
import numpy
from numpy import genfromtxt
import mxnet as mx
from mxnet import image, np, nd
from mxnet.gluon.data import dataset
from tivhelpfun import crop_scale, flip_horizontal, crop_scale_weight, flip_horizontal_weight


class TIVSegData(dataset.Dataset):
    """
    Usage: the user needs to provide a root directory that has the following structure:

        root:
            train:
                img/
                mask/
            val:
                img/
                mask/

                The images are writen as jpg and the mask as csv with entry the class number

    Parameters
    ----------
    root : directory
    image_shape : shape of image e.g. (3,480,480)
    mode : mode to load data and masks
    tiv_img_proc : augmentation mode
    cropW : crop the width of an image (Pixels)
    cropW : crop the height of an image (Pixels)
    img_width : width of original image
    img_height : height of original image
    max_num_img : maximum number of images per batch
    resize_mask_shape : resize shape of the input mask
    testing : flag for debugging

    """

    def __init__(self, root, image_shape, mode='train', tiv_img_proc='tiv', cropW=None, cropH=None, img_width=810, img_height=480, max_num_img = None, resize_mask_shape = None, testing = False):

        if tiv_img_proc is 'tiv' and mode is 'train':
            self._imgProc = ImageAugNormTIV(data_shape=image_shape, aug=True, norm=True)
        else:
            self._imgProc = ImageAugNormTIV(data_shape=image_shape, aug=False, norm=True)

        self._mode = mode
        self._cropW = cropW
        self._cropH = cropH
        self.preCalcMask = []
        self.testing = testing

        # Take into account how root directory is entered
        if root[-1] == '/':
            self._root = root
        else:
            self._root = root + '/'

        # Select the location to read images, different for train/val/test
        if self._mode == 'train':
            self._root_img = self._root + 'train/img/'
            self._root_mask = self._root + 'train/mask/'
        elif self._mode == 'val':
            self._root_img = self._root + 'val/img/'
            self._root_mask = self._root + 'val/mask/'
        else:
            raise Exception('modes avail: {train, val}')

        image_list = sorted(os.listdir(self._root_img))
        mask_list = []
        for fn in sorted(os.listdir(self._root_mask)):
            if fn.endswith('csv'):
                mask_list.append(fn)

        # Check input
        assert len(image_list) == len(mask_list), "Masks and labels have different count"

        # The original image is cut into three sub images with size cropW and cropH
        self._image_list = []
        for i in range(len(image_list)):
            self._image_list.append(
                ImageWithCrop(img_filepath=os.path.join(self._root_img, image_list[i]),
                              mask_filepath=os.path.join(self._root_mask, mask_list[i]), crop_w=self._cropW, crop_h=self._cropH,
                              crop_h_start=0, crop_w_start=0, resize_mask_shape = resize_mask_shape, grey_val=False))
            #self._image_list.append(
            #    ImageWithCrop(img_filepath=os.path.join(self._root_img, image_list[i]), mask_filepath=os.path.join(self._root_mask, mask_list[i]), crop_w=img_width, crop_h=img_height,
            #                  crop_w_start=(img_width/2 - self._cropW/2), crop_h_start=(img_height/2-self._cropH/2), grey_val=False))
            self._image_list.append(
                ImageWithCrop(img_filepath=os.path.join(self._root_img, image_list[i]), mask_filepath=os.path.join(self._root_mask, mask_list[i]), crop_w=img_width, crop_h=img_height,
                              crop_w_start=(img_width - self._cropW), crop_h_start=(img_height-self._cropH), grey_val=False))


        # Return only a maximum number of images per batch
        if max_num_img is not None and len(self._image_list) > max_num_img:
            self._max_num_img = max_num_img
            self._img_cropped_list = random.sample(range(0,len(self._image_list)),max_num_img)
        else:
            self._max_num_img = None
            self._img_cropped_list = []

        if cropW is not None:
            self._img_w = cropW
        else:
            self._img_w = img_width

        if cropH is not None:
            self._img_h = cropH
        else:
            self._img_h = img_height

    def __getitem__(self, idx):
        """
        Returns the image and mask for the requested index
        Parameters
        ----------
        idx : requested index
        """
        if self._max_num_img is not None:
            img_mask_getter = self._image_list[self._img_cropped_list[idx]]
        else:
            img_mask_getter = self._image_list[idx]

        img = img_mask_getter.get_image()
        mask = img_mask_getter.get_mask()

        img, mask = self._imgProc.aug_transform(data=img, mask=mask, img_h=self._img_h, img_w=self._img_w)
        base, mask = self._imgProc.norm_image_tiv(img=img, mask=mask)

        return base.astype(np.float32), mask.astype(np.float32)


    def __len__(self):
        """
        Returns the number of images for a batch
        Parameters
        ----------
        """
        if self._max_num_img is not None:
            return len(self._img_cropped_list)
        else:
            return len(self._image_list)

    def getitem_no_aug(self, idx):
        """
        Returns the image and mask for the requested index without augmentation
        Parameters
        ----------
        idx : requested index
        """
        img_mask_getter = self._image_list[idx]

        img = img_mask_getter.get_image()
        mask = img_mask_getter.get_mask()

        return img.astype(np.float32), mask.astype(np.float32)

    def get_size_img(self):
        """
        image size without augmentation
        Parameters
        ----------
        """
        img, mask = self.getitem_no_aug(0)
        return img.shape[0], img.shape[1]

class TIVSegData_Weight(dataset.Dataset):
    """
    Weighted database
    Usage: the user needs to provide a root directory that has the following structure:

        root:
            train:
                img/
                mask/
            val:
                img/
                mask/

                The images are writen as jpg and the mask as csv with entry the class number

    Parameters
    ----------
    root : directory
    image_shape : shape of image e.g. (3,480,480)
    mode : mode to load data and masks
    tiv_img_proc : augmentation mode
    cropW : crop the width of an image (Pixels)
    cropW : crop the height of an image (Pixels)
    img_width : width of original image
    img_height : height of original image
    max_num_img : maximum number of images per batch
    resize_mask_shape : resize shape of the input mask
    testing : flag for debugging
    """

    def __init__(self, root, image_shape, mode='train', tiv_img_proc='tiv', cropW=None, cropH=None, img_width=810, img_height=480, max_num_img = None, resize_mask_shape = None, testing = False):

        if tiv_img_proc is 'tiv' and mode is 'train':
            self._imgProc = ImageAugNormTIV_Weight(data_shape=image_shape, aug=True, norm=True)
        else:
            self._imgProc = ImageAugNormTIV_Weight(data_shape=image_shape, aug=False, norm=True)

        self._mode = mode  # train/val mode
        self._cropW = cropW
        self._cropH = cropH
        self.preCalcMask = []
        self.testing = testing

        # Take into account how root directory is entered
        if root[-1] == '/':
            self._root = root
        else:
            self._root = root + '/'

        # Select the location to read images, different for train/val/test
        if self._mode == 'train':
            self._root_img = self._root + 'train/img/'
            self._root_mask = self._root + 'train/mask/'
            self._root_weight = self._root + 'train/mask/'
        elif self._mode == 'val':
            self._root_img = self._root + 'val/img/'
            self._root_mask = self._root + 'val/mask/'
            self._root_weight = self._root + 'val/mask/'
        else:
            raise Exception('modes avail: {train, val}')

        image_list = sorted(os.listdir(self._root_img))
        mask_list = []
        weight_list = []
        for fn in sorted(os.listdir(self._root_mask)):
            if fn.endswith('csv'):
                mask_list.append(fn)
            if fn.endswith('.weight'):
                weight_list.append(fn)
        
        # Check input
        
        assert len(image_list) == len(mask_list) == len(weight_list), "Masks and labels have different count"+len(image_list)+' '+len(mask_list)+' '+len(weight_list)+' '

        self._image_list = []
        for i in range(len(image_list)):
            self._image_list.append(
                ImageWithCrop_Weight(img_filepath=os.path.join(self._root_img, image_list[i]),
                              mask_filepath=os.path.join(self._root_mask, mask_list[i]),
                              weight_filepath=os.path.join(self._root_weight, weight_list[i]),
                              crop_w=self._cropW, crop_h=self._cropH,
                              crop_h_start=0, crop_w_start=0, resize_mask_shape = resize_mask_shape, grey_val=False))
            #self._image_list.append(
            #    ImageWithCrop_Weight(img_filepath=os.path.join(self._root_img, image_list[i]),
            #                  mask_filepath=os.path.join(self._root_mask, mask_list[i]),
            #                  weight_filepath=os.path.join(self._root_weight, weight_list[i]),
            #                  crop_w=img_width, crop_h=img_height,
            #                  crop_w_start=(img_width/2 - self._cropW/2), crop_h_start=(img_height/2-self._cropH/2), grey_val=True))
            self._image_list.append(
                ImageWithCrop_Weight(img_filepath=os.path.join(self._root_img, image_list[i]),
                              mask_filepath=os.path.join(self._root_mask, mask_list[i]),
                              weight_filepath=os.path.join(self._root_weight, weight_list[i]),
                              crop_w=img_width, crop_h=img_height,
                              crop_w_start=(img_width - self._cropW), crop_h_start=(img_height-self._cropH), grey_val=False))

        if max_num_img is not None and len(self._image_list) > max_num_img:
            self._max_num_img = max_num_img
            self._img_cropped_list = random.sample(range(0,len(self._image_list)),max_num_img)
        else:
            self._max_num_img = None
            self._img_cropped_list = []

        if cropW is not None:
            self._img_w = cropW
        else:
            self._img_w = img_width

        if cropH is not None:
            self._img_h = cropH
        else:
            self._img_h = img_height

    def __getitem__(self, idx):
        """
        Returns the image and mask for the requested index
        Parameters
        ----------
        idx : requested index
        """
        if self._max_num_img is not None:
            img_mask_getter = self._image_list[self._img_cropped_list[idx]]
        else:
            img_mask_getter = self._image_list[idx]

        img = img_mask_getter.get_image()
        mask = img_mask_getter.get_mask()
        weight = img_mask_getter.get_weight()

        img, mask, weight = self._imgProc.aug_transform(data=img, mask=mask, weight=weight, img_h=self._img_h, img_w=self._img_w)
        base, mask, weight = self._imgProc.norm_image_tiv(img=img, mask=mask, weight=weight)

        return base.astype(np.float32), mask.astype(np.float32), weight.astype(np.float32)


    def __len__(self):
        """
        Returns the number of images per batch
        Parameters
        ----------
        """
        if self._max_num_img is not None:
            return len(self._img_cropped_list)
        else:
            return len(self._image_list)

    def getitem_no_aug(self, idx):
        """
        Returns the image and mask for the requested index without augmentation
        Parameters
        ----------
        idx : requested index
        """
        img_mask_getter = self._image_list[idx]

        img = img_mask_getter.get_image()
        mask = img_mask_getter.get_mask()
        weight = img_mask_getter.get_weight()

        return img.astype(np.float32), mask.astype(np.float32)

    def get_size_img(self):
        """
        image size without augmentation
        Parameters
        ----------
        """
        img, mask, weight = self.getitem_no_aug(0)
        return img.shape[0], img.shape[1]

class ImageWithCrop:
    """
    Image loader class with crop
    Parameters
    ----------
    img_filepath : path to the image
    mask_filepath : path to the mask
    crop_w : width of cropped image
    crop_h : height of cropped image
    crop_w_start : start point for cropping width
    crop_h_start : start point for cropping height
    resize_mask_shape : should mask be resized for augmentation
    image_name : special name
    grey_val : grey value image
    """
    def __init__(self, img_filepath, mask_filepath, crop_w=None, crop_h=None, crop_w_start=0, crop_h_start=0, resize_mask_shape = None, image_name=None, grey_val=True):
        self._img_filepath = img_filepath
        self._mask_filepath = mask_filepath
        self._cropW = crop_w
        self._cropH = crop_h
        self._cropW_start = crop_w_start
        self._cropH_start = crop_h_start
        self._resize_mask_shape = resize_mask_shape
        self.image_name = image_name
        self.grey_val = grey_val

    def get_image(self):
        # Image is read in an HWC
        img = image.imread(self._img_filepath)
        if self._cropW is not None:
            img = img[:, self._cropW_start:self._cropW, :]
        if self._cropH is not None:
            img = img[self._cropH_start:self._cropH, :, :]
        if self.grey_val:
            img = img[:,:,0]
        return img

    def get_mask(self):
        # Mask is read in as HW
        mask = genfromtxt(self._mask_filepath, delimiter=',')
        mask = mask[:, self._cropW_start:self._cropW]
        mask = mask[self._cropH_start:self._cropH, :]
        if self._resize_mask_shape is not None:
            mask = numpy.resize(mask, self._resize_mask_shape)
        return nd.array(mask)

class ImageWithCrop_Weight:
    """
    Image loader class with crop and wight
    Parameters
    ----------
    img_filepath : path to the image
    mask_filepath : path to the mask
    weight_filepath : path to the weight
    crop_w : width of cropped image
    crop_h : height of cropped image
    crop_w_start : start point for cropping width
    crop_h_start : start point for cropping height
    resize_mask_shape : should mask be resized for augmentation
    image_name : special name
    grey_val : grey value image
    """
    def __init__(self, img_filepath, mask_filepath, weight_filepath, crop_w=None, crop_h=None, crop_w_start=0, crop_h_start=0, resize_mask_shape = None, image_name=None, grey_val=True):
        self._img_filepath = img_filepath
        self._mask_filepath = mask_filepath
        self._weight_filepath = weight_filepath
        self._cropW = crop_w
        self._cropH = crop_h
        self._cropW_start = crop_w_start
        self._cropH_start = crop_h_start
        self._resize_mask_shape = resize_mask_shape
        self._resize_weight_shape = resize_mask_shape
        self.image_name = image_name
        self.grey_val = grey_val

    def get_image(self):
        # Image is read in an HWC
        img = image.imread(self._img_filepath)
        if self._cropW is not None:
            img = img[:, self._cropW_start:self._cropW, :]
        if self._cropH is not None:
            img = img[self._cropH_start:self._cropH, :, :]
        if self.grey_val:
            img = img[:,:,0]
        return img

    def get_mask(self):
        # Mask is read in as HW
        mask = genfromtxt(self._mask_filepath, delimiter=',')
        mask = mask[:, self._cropW_start:self._cropW]
        mask = mask[self._cropH_start:self._cropH, :]
        if self._resize_mask_shape is not None:
            mask = numpy.resize(mask, self._resize_mask_shape)
        return nd.array(mask)

    def get_weight(self):
        # Mask is read in as HW
        weight = genfromtxt(self._weight_filepath, delimiter=',')
        weight = weight[:, self._cropW_start:self._cropW]
        weight = weight[self._cropH_start:self._cropH, :]
        if self._resize_weight_shape is not None:
            weight = numpy.resize(weight, self._resize_weight_shape)
        return nd.array(weight)

class ImageAugNormTIV:
    """
    Augmentation and norm
    Parameters
    ----------
    data_shape : shape of the desired data
    aug : switch for augmentation
    norm : switch for normalization
    """
    def __init__(self, data_shape, aug=True, norm=True):
        self._aug = aug
        self._norm = norm
        self._data_shape = data_shape

    def aug_transform(self, data, mask, img_w, img_h ):
        # Normalize the image for augmentation
        if numpy.amax(data) >= 2:
            data = data.astype('float32') / 255
        if self._aug:
            augs = mx.image.CreateAugmenter(data_shape=self._data_shape, rand_mirror=True, rand_resize=True, rand_crop=True,
                                            inter_method=10, brightness=0.01, contrast=0.01, pca_noise=0.001)

            for aug in augs:
                if isinstance(aug, mx.image.image.HorizontalFlipAug):
                    data, mask = flip_horizontal(data, mask)
                elif isinstance(aug, mx.image.image.RandomSizedCropAug):
                    data, mask = crop_scale(data, mask, img_w=img_w, img_h=img_h)
                else:
                    data = aug(data)
            return data, mask
        else:
            return data, mask

    def norm_image_tiv(self, img, mask):
        # Normalize the image
        if self._norm:
            base = img.astype('float32')
            # NCHW
            if len(base.shape) == 2:
                base = nd.expand_dims(base, axis=2)
            base = nd.transpose(base, axes=(2, 0, 1))
            if numpy.amax(base) >= 2:
                base = base / 255.0

            return base, mask
        else:
            return img, mask

    def random_pad(vec, pad_width, *_, **__):
        #creates random padding
        vec[:pad_width[0]] = numpy.ones(int(pad_width[0]))  # numpy.random.randint(20, 30, size=pad_width[0])
        vec[vec.size - pad_width[1]:] = numpy.ones(int(pad_width[1]))  # numpy.random.randint(30, 40, size=pad_width[1])


class ImageAugNormTIV_Weight:
    """
    Augmentation and norm with weight
    Parameters
    ----------
    data_shape : shape of the desired data
    aug : switch for augmentation
    norm : switch for normalization
    """
    def __init__(self, data_shape, aug=True, norm=True):
        self._aug = aug
        self._norm = norm
        self._data_shape = data_shape

    def aug_transform(self, data, mask, weight, img_w, img_h):
        # Normalize the image for augmentation
        if numpy.amax(data) >= 2:
            data = data.astype('float32') / 255
        if self._aug:
            augs = mx.image.CreateAugmenter(data_shape=self._data_shape, rand_mirror=True, rand_resize=True, rand_crop=True,
                                            inter_method=10, brightness=0.01, contrast=0.01, pca_noise=0.001)


            for aug in augs:
                if isinstance(aug, mx.image.image.HorizontalFlipAug):
                    data, mask, weight = flip_horizontal_weight(data, mask, weight)
                elif isinstance(aug, mx.image.image.RandomSizedCropAug):
                    data, mask, weight = crop_scale_weight(data, mask, weight, img_w=img_w, img_h=img_h)
                else:
                    data = aug(data)
            return data, mask, weight
        else:
            return data, mask, weight

    def norm_image_tiv(self, img, mask, weight):
        #normalize the data
        if self._norm:
            base = img.astype('float32')
            # NCHW
            if len(base.shape) == 2:
                base = nd.expand_dims(base, axis=2)
            base = nd.transpose(base, axes=(2, 0, 1))
            if numpy.amax(base) >= 2:
                base = base / 255.0

            return base, mask, weight
        else:
            return img, mask, weight

    def random_pad(vec, pad_width, *_, **__):
        # creates random padding
        vec[:pad_width[0]] = numpy.ones(int(pad_width[0]))  # numpy.random.randint(20, 30, size=pad_width[0])
        vec[vec.size - pad_width[1]:] = numpy.ones(int(pad_width[1]))  # numpy.random.randint(30, 40, size=pad_width[1])
