'''
Copyright (c) 2019-2021 TIV Consultancy. All right reserved

No redistribution in any form allowed.
'''

import random
import numpy
import cv2
import mxnet as mx
from mxnet import np, nd, npx
from matplotlib import pyplot as plt


def bilinear_kernel(in_channels, out_channels, kernel_size):
    """
    Kerner for bilinear interpolation
    """
    factor = (kernel_size + 1) // 2
    if kernel_size % 2 == 1:
        center = factor - 1
    else:
        center = factor - 0.5
    og = (np.arange(kernel_size).reshape(-1, 1),
          np.arange(kernel_size).reshape(1, -1))
    filt = (1 - np.abs(og[0] - center) / factor) * \
           (1 - np.abs(og[1] - center) / factor)
    weight = np.zeros((in_channels, out_channels, kernel_size, kernel_size))
    weight[range(in_channels), range(out_channels), :, :] = filt
    return np.array(weight)


def plot_mx_array(array):
    """
    Plots array.
    Array expected to be height x width x 3 (channels), and values are floats between 0 and 255.
    """
    # assert array.shape[2] == 3 or array.shape[2] == 1, "RGB Channel should be last"
    plt.imshow(array.asnumpy())
    plt.show()


def flip_horizontal(data, mask):
    """
    Flips the input horizontal for augmentation
    """
    rand_hf = random.random()
    if rand_hf > 0.5:
        flip_hori = mx.image.image.HorizontalFlipAug(p=1)
        mask = flip_hori(mask)
        data = flip_hori(data)
    return data, mask


def crop_scale(data, mask, p_activation=0.66, p_scale=0.5, p_rand_pad=0.5, img_h=320, img_w=480):
    """
    Random cropping, padding and scaling of the input data, chance for cropping, scaling and padding can controlled through
     parameters
    """
    rand_scale = random.random()
    if rand_scale < p_activation:
        minsize_h = img_h//2
        maxsize_h = img_h
        minsize_w = img_w//2
        maxsize_w = img_w
        size_h = minsize_h + int(random.random() * (maxsize_h - minsize_h))
        size_w = minsize_w + int(random.random() * (maxsize_w - minsize_w))
        start_h = int(random.random() * (maxsize_h - size_h))
        start_w = int(random.random() * (maxsize_w - size_w))
        mask = mask[start_h:start_h + size_h, start_w:start_w + size_w]
        data = data[start_h:start_h + size_h, start_w:start_w + size_w, :]
        # mask = mx.image.fixed_crop(mask,startW,startH,startW+sizeW,startH+sizeH)
        # data = mx.image.fixed_crop(data,startW,startH,startW+sizeW,startH+sizeH)
        if random.random() < p_scale:
            space_left = int(random.random() * (maxsize_w - size_w))
            space_top = int(random.random() * (maxsize_h - size_h))
            mask = numpy.pad(mask.asnumpy(),
                             ((space_top, maxsize_h - size_h - space_top),
                              (space_left, maxsize_w - size_w - space_left)))
            if random.random() < p_rand_pad:
                data = numpy.pad(data.asnumpy(), (
                    (space_top, maxsize_h - size_h - space_top), (space_left, maxsize_w - size_w - space_left), (0, 0)))
                #for i in range(len(data)):
                #    for j in range(len(data[i])):
                #        rand = random.random()
                #        for t in range(len(data[i][j])):
                #            if data[i][j][t] == 0:
                #                data[i][j][t] = rand
            else:
                data = numpy.pad(data.asnumpy(), (
                    (space_top, maxsize_h - size_h - space_top), (space_left, maxsize_w - size_w - space_left), (0, 0)))
                #for i in range(len(data)):
                #    for j in range(len(data[i])):
                #        rand = random.random()
                #        for t in range(len(data[i][j])):
                #            if data[i][j][t] == 0:
                #                data[i][j][t] = rand

        else:
            mask = cv2.resize(mask.asnumpy(), (img_w, img_h))
            data = cv2.resize(data.asnumpy(), (img_w, img_h))
        mask = nd.array(mask)
        data = nd.array(data)
        # mask = aug(mask)
        # data = aug(data)

    return data, mask


def accuracy_tiv(data, mask, axis=1):
    # type: (mx.ndarray, mx.ndarray, int) -> float
    # accurcay between data and mask with softmax
    diff = mask - nd.argmax(nd.softmax(data, axis=axis), axis=axis)
    accu = nd.sum(nd.abs(diff)) / diff.size
    return accu.asnumpy().item()


def accuracy_dice_tiv(data, mask, axis=1):
    # type: (mx.ndarray, mx.ndarray, int) -> float
    # accurcay between data and mask with Dice
    if type(data) is type(()):
        data = data[0]
    pred = nd.argmax(nd.softmax(data, axis=axis), axis=axis)
    tp = nd.sum(nd.where(mask, mask == pred, nd.zeros_like(mask)))
    fn = nd.sum(nd.where(mask, mask != pred, nd.zeros_like(mask)))
    fp = nd.sum(nd.where(pred, mask != pred, nd.zeros_like(mask)))
    accu = 2 * tp / (2 * tp + fp + fn)
    accu = accu.astype('double', copy=True)
    tp = tp.astype('double', copy=True)
    fn = fn.astype('double', copy=True)
    fp = fp.astype('double', copy=True)
    return accu.asnumpy().item(), tp.asnumpy().item(), fn.asnumpy().item(), fp.asnumpy().item()


def validation(net, data_val, ctx, fun=accuracy_dice_tiv):
    # type: (mxnet.block, mxnet.gluon.image_input.DataLoader, mxnet.context) -> float
    # Validate a net
    import time
    count = 0
    time_per_picture = 0
    for i, (data, target) in enumerate(data_val):
        start = time.time();
        data_gpu = data.copyto(ctx)
        target_gpu = target.copyto(ctx)
        outputs = net(data_gpu)
        if i is 0:
            accumulator = fun(outputs, target_gpu, axis=1)
        else:
            accumulator += fun(outputs, target_gpu, axis=1)
        count += 1
        time_per_picture += time.time() - start
        print(data.shape)
        print(str(time_per_picture/count/data.shape[0]))
    return numpy.asanyarray(accumulator) / count


def crop_scale_weight(data, mask, weight, img_h, img_w, p_activation=0.66, p_scale=0.5, p_rand_pad=0.5, ):
    """
    Random cropping, padding and scaling of the input data, chance for cropping, scaling and padding can controlled through
     parameters
    """
    rand_scale = random.random()
    if rand_scale < p_activation:
        minsize_h = img_h//2
        maxsize_h = img_h
        minsize_w = img_w//2
        maxsize_w = img_w
        size_h = minsize_h + int(random.random() * (maxsize_h - minsize_h))
        size_w = minsize_w + int(random.random() * (maxsize_w - minsize_w))
        start_h = int(random.random() * (maxsize_h - size_h))
        start_w = int(random.random() * (maxsize_w - size_w))
        mask = mask[start_h:start_h + size_h, start_w:start_w + size_w]
        data = data[start_h:start_h + size_h, start_w:start_w + size_w, :]
        weight = weight[start_h:start_h + size_h, start_w:start_w + size_w]
        # mask = mx.image.fixed_crop(mask,startW,startH,startW+sizeW,startH+sizeH)
        # data = mx.image.fixed_crop(data,startW,startH,startW+sizeW,startH+sizeH)
        if random.random() < p_scale:
            space_left = int(random.random() * (maxsize_w - size_w))
            space_top = int(random.random() * (maxsize_h - size_h))
            mask = numpy.pad(mask.asnumpy(),
                             ((space_top, maxsize_h - size_h - space_top),
                              (space_left, maxsize_w - size_w - space_left)))

            weight = numpy.pad(weight.asnumpy(),
                             ((space_top, maxsize_h - size_h - space_top),
                              (space_left, maxsize_w - size_w - space_left)))
            if random.random() < p_rand_pad:
                data = numpy.pad(data.asnumpy(), (
                    (space_top, maxsize_h - size_h - space_top), (space_left, maxsize_w - size_w - space_left), (0, 0)))
                #for i in range(len(data)):
                #    for j in range(len(data[i])):
                #        rand = random.random()
                #        for t in range(len(data[i][j])):
                #            if data[i][j][t] == 0:
                #                data[i][j][t] = rand
            else:
                data = numpy.pad(data.asnumpy(), (
                    (space_top, maxsize_h - size_h - space_top), (space_left, maxsize_w - size_w - space_left), (0, 0)))
                #for i in range(len(data)):
                #    for j in range(len(data[i])):
                #        rand = random.random()
                #        for t in range(len(data[i][j])):
                #            if data[i][j][t] == 0:
                #                data[i][j][t] = rand

        else:
            mask = cv2.resize(mask.asnumpy(), (img_w, img_h))
            data = cv2.resize(data.asnumpy(), (img_w, img_h))
            weight = cv2.resize(weight.asnumpy(), (img_w, img_h))
        mask = nd.array(mask)
        data = nd.array(data)
        weight = nd.array(weight)
        # mask = aug(mask)
        # data = aug(data)


    return data, mask, weight


def flip_horizontal_weight(data, mask, weight):
    """
    Flips the input data
    """
    rand_hf = random.random()
    if rand_hf > 0.5:
        flip_hori = mx.image.image.HorizontalFlipAug(p=1)
        mask = flip_hori(mask)
        data = flip_hori(data)
        weight = flip_hori(weight)
    return data, mask, weight
